import shutil
from collections import Callable
from pathlib import Path
from typing import Dict

import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets, QtCore
from pyqtgraph.console import ConsoleWidget
from utilitys import ParamEditor, PrjParam
from utilitys.fns import attemptFileLoad
from utilitys.widgets import EasyWidget

from cubecorrections.rotation import RotateWindow
from s3a import S3A
from s3a.plugins.file import ProjectData
from s3a.processing import ImageProcess
from .helpers import FRPgParamDelegate
from .layermodel import LayerModel
from .layerview import LayerView
from .view3d import PCBViewer3D

pg.setConfigOption('imageAxisOrder', 'row-major')

#def find_circles(image: np.ndarray, fgVerts, radius=10, param2='test'):

def find_circles(func: Callable,image=None):

  # out = np.zeros(image.shape[:2], 'uint8')
  # for vert in fgVerts:
  #   cv.circle(out, tuple(vert), radius, 1, -1)
  # return ProcessIO(image=out > 0)

  def findCirclesProcessor():
    proc = ImageProcess()
    proc.addFunction(func)
    return proc
  #procCtor =  ImageProcess.fromFunction(find_circles)
  # PRJ_SINGLETON.imgProcClctn.addProcessCtor(findCirclesProcessor)


class MainWin(QtWidgets.QMainWindow):
  def __init__(self, parent: QtWidgets.QWidget=None, layerInfo: Dict[str, list]=None, dataCube=None):
    super().__init__(parent)
    self.toolsEditor = ParamEditor.buildClsToolsEditor(MainWin, 'Tools')

    from .algorithms import inputImage
    find_circles(inputImage)
    here = Path(__file__).parent
    projDir = here.parent/'xray-s3a-proj'
    if projDir.exists():
      shutil.rmtree(projDir/'images', ignore_errors=True)
      shutil.rmtree(projDir/'annotations', ignore_errors=True)
    else:
      ProjectData.create(name=projDir, cfg=attemptFileLoad(here.parent/'projcfg.s3aprj'))
    proj = ProjectData(projDir/'xray-s3a-proj.s3aprj')
    self.s3a = S3A(loadLastState=False, project=proj.cfgFname)

    self.model = LayerModel(dataCube=dataCube, layerInfo=layerInfo)
    self.viewer = LayerView(self.model)
    self.viewer_3d = PCBViewer3D(dataCube)
    # self.model.sigCubeChanged.connect(lambda: self.viewer_3d.setDataCube(
    #   self.model.dataCube))

    self.tbl = QtWidgets.QTableView()
    self.tbl.setModel(self.model.tblModel)
    self.tbl.selectionModel().selectionChanged.connect(self.updateTblViews)
    pgDict = dict(type='list', values={'True': True, 'False': False})
    self.tbl.setItemDelegateForColumn(0, FRPgParamDelegate(pgDict))
    self.tbl.setSelectionBehavior(self.tbl.SelectRows)

    self.plt = pg.PlotWidget()
    self.viewer.itemParent = self.plt
    for item in self.viewer.items:
      self.plt.addItem(item)
    vb = self.plt.getViewBox()
    vb.setAspectLocked(True)

    self.toolsEditor.registerFunc(self.fixRotation, name='Interactive Rotation Fix')
    self.toolsEditor.registerFunc(self.showDevConsole, name='Dev Console')
    self.toolsEditor.registerFunc(self.openS3A, name='Annotate Selected Layer Data')
    _, param = self.toolsEditor.registerFunc(self.highlight3DSlices, name='Show '
                                                                          'Selection in 3D',
                                  returnParam=True)
    MainWin.liveUpdate = self.toolsEditor.registerProp(PrjParam('Live Update 3D', False),
                                                       overrideBasePath=())
    # self.toolsEditor.registerFunc(self.viewer_3d.resetZoom())
    def maybeUpdate():
      if self.liveUpdate:
        self.highlight3DSlices(param['downsampleFactor'])
    self.tbl.selectionModel().selectionChanged.connect(maybeUpdate)
    def maybeShowMetric():
      if self.model.metricInfo is not self.model.NO_METRIC:
        self.model.metricInfo.show()
      else:
        QtWidgets.QMessageBox.information(self, 'No Metric', 'No metric to show.')
    self.toolsEditor.registerFunc(maybeShowMetric, name='Show Layer Slice Metrics')
    self._buildUi()

  def _buildUi(self):
    self.addTabbedDock(QtCore.Qt.LeftDockWidgetArea, self.toolsEditor.dock)
    menu = QtWidgets.QMenuBar(self)
    self.toolsEditor.show()
    self.setMenuBar(menu)

    EasyWidget.buildMainWin(
      [
        EasyWidget([self.viewer_3d, self.tbl], useSplitter=True, layout='V'),
        self.plt
      ], layout='H', win=self, useSplitter=True)

  def openS3A(self):
    layerDf = self.model.layerDf
    prj = self.s3a.filePlg.projData
    for ii, name in enumerate(layerDf['Layer Name']):
      imgName = prj.imagesDir/f'{name}.png'
      if imgName not in prj.images:
        selectedSlices = layerDf.loc[ii, 'Slices']
        img = self.model.dataCube[...,selectedSlices].mean(2).astype('uint8')
        prj.addImage(imgName, data=img, copyToProj=True, allowOverwrite=True)
    QtCore.QTimer.singleShot(0, self.s3a.showMaximized)
    self.s3a.filePlg._projImgMgr.show()

  def fixRotation(self):
    win = RotateWindow(self.model.dataCube)
    win.show()
    win.sigChangesAccepted.connect(lambda: self.model.setCube(win.cube))


  def showDevConsole(self):
    namespace = dict(win=self)
    # "dict" default is to use repr instead of string for internal elements, so expanding
    # into string here ensures repr is not used
    nsPrintout = [f"{k}: {v}" for k, v in namespace.items()]
    text = f'Starting console with variables:\n' \
           f'{nsPrintout}'
    console = ConsoleWidget(self, namespace=namespace, text=text)
    console.setWindowFlags(QtCore.Qt.Window)
    console.show()

  def updateTblViews(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
    selRows = [sel.row() for sel in selected.indexes()]
    deselRows = [sel.row() for sel in deselected.indexes()]
    selRows = np.unique(selRows)
    deselRows = np.unique(deselRows)
    self.model.tblModel.dataDf.loc[deselRows, 'View'] = False
    self.model.tblModel.dataDf.loc[selRows, 'View'] = True
    self.model.tblModel.sigDataChanged.emit()

  def highlight3DSlices(self, downsampleFactor=3):
    """
    :param downsampleFactor:
      limits: [1, 1000]
    """
    selectedRows = self.model.layerDf['View']
    if np.any(selectedRows):
      selectedSlices = np.concatenate(self.model.layerDf.loc[selectedRows, 'Slices'].values)
    else:
      selectedSlices = None
    self.viewer_3d.setDataCube(self.model.dataCube, selectedSlices, downsampleFactor)

  #noinspection PyTypeChecker
  def addTabbedDock(self, area: QtCore.Qt.DockWidgetArea, dockwidget: QtWidgets.QDockWidget):
    curAreaWidgets = [d for d in self.findChildren(QtWidgets.QDockWidget)
                      if self.dockWidgetArea(d) == area]
    try:
      self.tabifyDockWidget(curAreaWidgets[-1], dockwidget)
    except IndexError:
      # First dock in area
      self.addDockWidget(area, dockwidget)